﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

class Person {
	public string FirstName { get; set; }
	public string LastName { get; set; }

	public override string ToString() {
		return FirstName + " " + LastName;
	}
}

class Program {

	static void PrintList(List<object> list) {
		foreach (var item in list) {
			Console.WriteLine(item);
		}
	}

	static void PrintEnumerable(IEnumerable<object> list) {
		foreach (var item in list) {
			Console.WriteLine(item);
		}
	}

	static void Main(string[] args) {

		var people = new List<Person> {
			new Person { FirstName = "Pavel", LastName = "Jezek"},
			new Person { FirstName = "Petr", LastName = "Hnetynka"},
			new Person { FirstName = "Tom", LastName = "Holan"}
		};

		// Error: Argument 1: cannot convert from 'System.Collections.Generic.List<Person>' to 'System.Collections.Generic.List<object>'
		// PrintList(people);

		PrintEnumerable(people);

		var numbers = new List<byte> { 1, 2, 3 };
		// Error: Argument 1: cannot convert from 'System.Collections.Generic.List<byte>' to 'System.Collections.Generic.IEnumerable<object>'
		// Why? Interfaces are variant only for reference types (as arrays are)!
		// PrintEnumerable(numbers);

		// But why are interfaces variant only for reference types?
		// Let's say, the call above would compile:
		// 1) list in PrintEnumerable point to -> List<byte>
		// 2) foreach => GetEnumerator() => List<byte>.Enumerator
		// 3) list is IEnumerable<object> => item is object (static type)
		// 4) List<byte>.Enumerator.Current returns byte (static type)
		// 5) object item = (byte) List<byte>.Enumerator.Current requires boxing (value type into ref. variable) = special CIL instruction needed
		// => different CIL implementations of PrintEnumerable depending on the value passed:
		//    a) for list == IEnum<T> where T : struct MUST contain box CIL instruction
		//	  b) for list == IEnum<T> where T : class CANNOT contain box CIL instruction
		//    => impossible!

		DoTest();
	}

	static void DoTest() {
		IEnumerable<object[]> list1 = new List<string[]>();

		// Error: Cannot implicitly convert type 'System.Collections.Generic.List<int[]>' to 'System.Collections.Generic.IEnumerable<object[]>'.An explicit conversion exists (are you missing a cast?)
		// IEnumerable<object[]> list2 = new List<int[]> { new int[] { 1, 2, 3 } };

		IEnumerable<object[]> list2 = (IEnumerable<object[]>) new List<int[]> { new int[] { 1, 2, 3 } };
		foreach (var a in list2) {
			Console.WriteLine($"Got array: {a}");
			foreach (var o in a) {
				Console.WriteLine($"  Got object: {o}");
			}
		}

		// object[] a = (object[]) new int[] { 1, 2, 3 };
	}
}

